import ast
import inspect
import textwrap
from pathlib import Path
from dbgpy import config
from warnings import warn


def get_frame_above():
    frame = inspect.currentframe()
    outer_frame = inspect.getouterframes(frame)[2][
        0
    ]  # Move two up the stack since this is called from the child frame
    return outer_frame


def _get_func_obj(frame):
    """
    Get function object from a frame. If it can't find it, return None
    """

    codename = frame.f_code.co_name
    fobj = None

    try:
        if "self" in frame.f_locals:  # regular method
            fobj = getattr(frame.f_locals["self"].__class__, codename)
        elif "cls" in frame.f_locals:  # class method
            fobj = getattr(frame.f_locals["cls"], codename)
        else:
            module = inspect.getmodule(frame)  # only fetch module if we need it

            if hasattr(module, codename):  # regular module level function
                fobj = getattr(module, codename)
            else:  # static method
                classes = [
                    getattr(module, name)
                    for name in dir(module)
                    if inspect.isclass(getattr(module, name))
                ]
                for cls in classes:
                    if (
                        hasattr(cls, codename)
                        and getattr(cls, codename).__code__ == frame.f_code
                    ):
                        fobj = getattr(cls, codename)
                        break
        if fobj is None:
            """it's likely some nested function/method or a lambda, who logs in a lambda?"""
            return None
        return fobj
    except Exception:
        """never break logging"""
        return None


def _get_call_args_from_frame(frame):
    """Get AST nodes for each argument in the function call."""
    frame_info = inspect.getframeinfo(frame)

    # Line number of this function call
    line_number = frame_info.lineno

    # Line number of parent function definition
    parent_line_number = frame.f_code.co_firstlineno

    # Get the function calling this function
    calling_function = _get_func_obj(frame)

    # Parse the function to an ast
    calling_source, function_ast = _parse_ast(calling_function, frame)
    if calling_source is None and function_ast is None:
        return None, None, None

    # Line number of this function call
    function_call_line_number = line_number - parent_line_number + 1

    # Find the call object in the function ast matching the line number
    call_object = None
    for node in ast.walk(function_ast):
        if isinstance(node, ast.Expr) and isinstance(node.value, ast.Call):
            if node.lineno == function_call_line_number:
                call_object = node.value
                break
    # Get the call object's arguments
    call_args = call_object.args

    return call_args, calling_source, parent_line_number - 1


def _parse_ast(calling_function, frame):
    if frame.f_code.co_filename == '<stdin>':
        warn('Cannot get source from stdin (for example in Python REPL), dbg will default to printing normally')
        return None, None
    elif '<cell line: ' in frame.f_code.co_name:
        warn('dbg currently doesn\'t support IPython cells, dbg will default to printing normally')
        return None, None
    elif calling_function is None:
        # In a module
        calling_source = inspect.getsource(frame)
        function_ast = ast.parse(calling_source)
    else:
        # In a function we just need to parse the function
        calling_source = inspect.getsource(calling_function)
        # Unindent the source
        calling_source = textwrap.dedent(calling_source)
        function_ast = ast.parse(calling_source)
    return calling_source, function_ast


def frame_file_path(frame):
    """
    Return the path of the frame
    """
    try:
        if config.project_path:
            return Path(frame.f_code.co_filename).relative_to(
                Path.cwd()
            )
        else:
            return Path(frame.f_code.co_filename).name
    except ValueError:
        return '<unknown>'



from attr import attrs

__all__ = ["config"]


@attrs(auto_attribs=True)
class _Config:
    """Global mutable configuration object for dbgpy"""

    project_path: bool = True
    """Print the file path relative to the project root"""

    prefix_format: str = "{path}:"
    """Format string for the prefix of the debug output. `path` is the path of the file."""

    return_result: bool = True
    """Return the result of the expressions after printing"""


config = _Config()



from typing import Any, List, Optional
import ast

from dbgpy._stack_inspect import (
    get_frame_above,
    _get_call_args_from_frame,
    frame_file_path,
)
from dbgpy import config

try:
    import torch
except ImportError:
    torch = None

try:
    import astunparse
except ImportError:
    astunparse = None

__all__ = ["dbg"]


def _space_all_but_first(s: str, n_spaces: int) -> str:
    """Pad all lines except the first with n_spaces spaces"""
    lines = s.splitlines()
    for i in range(1, len(lines)):
        lines[i] = " " * n_spaces + lines[i]
    return "\n".join(lines)


def _format_value(val: Any, _repr=False):
    # Guards for special types
    if torch is not None and isinstance(val, torch.Tensor) and not _repr:
        val_string = str(val)
        end_of_tensor = val_string.rfind(")")
        new_val_string = f"{val_string[:end_of_tensor]}, shape={val.shape}, {val.device}{val_string[end_of_tensor:]}"
        return new_val_string

    # Otherwise, default to repr or str
    if _repr or isinstance(val, str):
        return repr(val)
    return str(val)


def _print_spaced(prefix: str, val_str: str, **kwargs):
    """Print variables with their variable names"""
    n_spaces = len(prefix)
    spaced = _space_all_but_first(val_str, n_spaces)
    print(
        f"{prefix} {spaced}",
        **kwargs,
    )


def _format_prefix(
        file_path: str, lineno: int, expression_string: Optional[str], val: Any
):
    inner_path = f"{file_path}:{lineno}"
    path = config.prefix_format.format(path=inner_path)

    # Use expression only if expression is not a literal string
    return f"{path} {expression_string} =" if expression_string else f"{path}"


def _get_expression(call_source_lines: List[str], arg) -> Optional[str]:
    if isinstance(arg, (ast.Constant, ast.JoinedStr)):
        return None

    # If multiline expression, clean up using ast.unparse (or astunparse)
    if arg.lineno != arg.end_lineno:
        if hasattr(ast, "unparse"):  # If in version 3.9 or newer
            return ast.unparse(arg)
        else:  # Otherwise, use astunparse
            return astunparse.unparse(arg)

    # Otherwise, just return the literal expression string
    return call_source_lines[arg.lineno - 1][
           arg.col_offset: arg.end_col_offset
           ].strip()


def return_vals(vals):
    if not config.return_result:
        return None
    if len(vals) == 1:
        return vals[0]
    return vals


def dbg(*vals, **kwargs):
    """
    Debug output with expression and value
    """
    outer_frame = get_frame_above()

    # Parse the stack frame, return source and arg nodes
    call_args, call_source, base_lineno = _get_call_args_from_frame(outer_frame)
    if call_args is None:
        print(*vals)
        return return_vals(vals)
    call_source_lines = call_source.splitlines()

    assert len(call_args) == len(vals)

    for arg, val in zip(call_args, vals):
        expression = _get_expression(call_source_lines, arg)
        prefix = _format_prefix(
            frame_file_path(outer_frame), arg.lineno + base_lineno, expression, val
        )
        val_str = _format_value(val)
        _print_spaced(prefix, val_str)

    return return_vals(vals)
